<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\ActivityBundle\Export\Export;

use Chill\ActivityBundle\Entity\ActivityReason;
use Chill\ActivityBundle\Repository\ActivityRepository;
use Chill\ActivityBundle\Security\Authorization\ActivityStatsVoter;
use Chill\MainBundle\Export\FormatterInterface;
use Chill\MainBundle\Export\ListInterface;
use Chill\MainBundle\Templating\TranslatableStringHelperInterface;
use DateTime;
use Doctrine\DBAL\Exception\InvalidArgumentException;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Query;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Security\Core\Role\Role;
use Symfony\Component\Validator\Constraints\Callback;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use Symfony\Contracts\Translation\TranslatorInterface;

use function array_key_exists;
use function count;
use function in_array;

class ListActivity implements ListInterface
{
    protected EntityManagerInterface $entityManager;

    protected array $fields = [
        'id',
        'date',
        'durationTime',
        'attendee',
        'list_reasons',
        'user_username',
        'circle_name',
        'type_name',
        'person_firstname',
        'person_lastname',
        'person_id',
    ];

    protected TranslatableStringHelperInterface $translatableStringHelper;

    protected TranslatorInterface $translator;

    private ActivityRepository $activityRepository;

    public function __construct(
        EntityManagerInterface $em,
        TranslatorInterface $translator,
        TranslatableStringHelperInterface $translatableStringHelper,
        ActivityRepository $activityRepository
    ) {
        $this->entityManager = $em;
        $this->translator = $translator;
        $this->translatableStringHelper = $translatableStringHelper;
        $this->activityRepository = $activityRepository;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder->add('fields', ChoiceType::class, [
            'multiple' => true,
            'expanded' => true,
            'choices' => array_combine($this->fields, $this->fields),
            'label' => 'Fields to include in export',
            'constraints' => [new Callback([
                'callback' => static function ($selected, ExecutionContextInterface $context) {
                    if (count($selected) === 0) {
                        $context->buildViolation('You must select at least one element')
                            ->atPath('fields')
                            ->addViolation();
                    }
                },
            ])],
        ]);
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_LIST];
    }

    public function getDescription()
    {
        return 'List activities';
    }

    public function getLabels($key, array $values, $data)
    {
        switch ($key) {
            case 'date':
                return static function ($value) {
                    if ('_header' === $value) {
                        return 'date';
                    }

                    $date = DateTime::createFromFormat('Y-m-d H:i:s', $value);

                    return $date->format('d-m-Y');
                };

            case 'attendee':
                return static function ($value) {
                    if ('_header' === $value) {
                        return 'attendee';
                    }

                    return $value ? 1 : 0;
                };

            case 'list_reasons':
                $activityRepository = $this->activityRepository;

                return function ($value) use ($activityRepository): string {
                    if ('_header' === $value) {
                        return 'activity reasons';
                    }

                    $activity = $activityRepository->find($value);

                    return implode(', ', array_map(function (ActivityReason $r) {
                        return '"' .
                                $this->translatableStringHelper->localize($r->getCategory()->getName())
                                . ' > ' .
                                $this->translatableStringHelper->localize($r->getName())
                                . '"';
                    }, $activity->getReasons()->toArray()));
                };

            case 'circle_name':
                return function ($value) {
                    if ('_header' === $value) {
                        return 'circle';
                    }

                    return $this->translatableStringHelper->localize(json_decode($value, true));
                };

            case 'type_name':
                return function ($value) {
                    if ('_header' === $value) {
                        return 'activity type';
                    }

                    return $this->translatableStringHelper->localize(json_decode($value, true));
                };

            default:
                return static function ($value) use ($key) {
                    if ('_header' === $value) {
                        return $key;
                    }

                    return $value;
                };
        }
    }

    public function getQueryKeys($data)
    {
        return $data['fields'];
    }

    public function getResult($query, $data)
    {
        return $query->getQuery()->getResult(Query::HYDRATE_SCALAR);
    }

    public function getTitle()
    {
        return 'List activities';
    }

    public function getType()
    {
        return 'activity';
    }

    public function initiateQuery(array $requiredModifiers, array $acl, array $data = [])
    {
        $centers = array_map(static function ($el) {
            return $el['center'];
        }, $acl);

        // throw an error if any fields are present
        if (!array_key_exists('fields', $data)) {
            throw new InvalidArgumentException('Any fields have been checked.');
        }

        $qb = $this->entityManager->createQueryBuilder();

        $qb
            ->from('ChillActivityBundle:Activity', 'activity')
            ->join('activity.person', 'person')
            ->join('person.center', 'center')
            ->andWhere('center IN (:authorized_centers)')
            ->setParameter('authorized_centers', $centers);

        foreach ($this->fields as $f) {
            if (in_array($f, $data['fields'], true)) {
                switch ($f) {
                    case 'id':
                        $qb->addSelect('activity.id AS id');

                        break;

                    case 'person_firstname':
                        $qb->addSelect('person.firstName AS person_firstname');

                        break;

                    case 'person_lastname':
                        $qb->addSelect('person.lastName AS person_lastname');

                        break;

                    case 'person_id':
                        $qb->addSelect('person.id AS person_id');

                        break;

                    case 'user_username':
                        $qb->join('activity.user', 'user');
                        $qb->addSelect('user.username AS user_username');

                        break;

                    case 'circle_name':
                        $qb->join('activity.scope', 'circle');
                        $qb->addSelect('circle.name AS circle_name');

                        break;

                    case 'type_name':
                        $qb->join('activity.type', 'type');
                        $qb->addSelect('type.name AS type_name');

                        break;

                    case 'list_reasons':
                        // this is a trick... The reasons is filled with the
                        // activity id which will be used to load reasons
                        $qb->addSelect('activity.id AS list_reasons');

                        break;

                    default:
                        $qb->addSelect(sprintf('activity.%s as %s', $f, $f));

                        break;
                }
            }
        }

        return $qb;
    }

    public function requiredRole()
    {
        return new Role(ActivityStatsVoter::LISTS);
    }

    public function supportsModifiers()
    {
        return ['activity', 'person'];
    }
}
