<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\ActivityBundle\Export\Export;

use Chill\ActivityBundle\Repository\ActivityRepository;
use Chill\ActivityBundle\Security\Authorization\ActivityStatsVoter;
use Chill\MainBundle\Export\ExportInterface;
use Chill\MainBundle\Export\FormatterInterface;
use Doctrine\ORM\Query;
use LogicException;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Security\Core\Role\Role;

/**
 * This export allow to compute stats on activity duration.
 *
 * The desired stat must be given in constructor.
 */
class StatActivityDuration implements ExportInterface
{
    public const SUM = 'sum';

    /**
     * The action for this report.
     */
    protected string $action;

    private ActivityRepository $activityRepository;

    /**
     * @param string $action the stat to perform
     */
    public function __construct(
        ActivityRepository $activityRepository,
        string $action = 'sum'
    ) {
        $this->action = $action;
        $this->activityRepository = $activityRepository;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_TABULAR];
    }

    public function getDescription()
    {
        if (self::SUM === $this->action) {
            return 'Sum activities duration by various parameters.';
        }
    }

    public function getLabels($key, array $values, $data)
    {
        if ('export_stat_activity' !== $key) {
            throw new LogicException(sprintf('The key %s is not used by this export', $key));
        }

        $header = self::SUM === $this->action ? 'Sum of activities duration' : false;

        return static fn (string $value) => '_header' === $value ? $header : $value;
    }

    public function getQueryKeys($data)
    {
        return ['export_stat_activity'];
    }

    public function getResult($qb, $data)
    {
        return $qb->getQuery()->getResult(Query::HYDRATE_SCALAR);
    }

    public function getTitle()
    {
        if (self::SUM === $this->action) {
            return 'Sum activity duration';
        }
    }

    public function getType()
    {
        return 'activity';
    }

    public function initiateQuery(array $requiredModifiers, array $acl, array $data = [])
    {
        $centers = array_map(
            static fn (array $el): string => $el['center'],
            $acl
        );

        $qb = $this->activityRepository->createQueryBuilder('activity');

        $select = null;

        if (self::SUM === $this->action) {
            $select = 'SUM(activity.durationTime) AS export_stat_activity';
        }

        return $qb->select($select)
            ->join('activity.person', 'person')
            ->join('person.center', 'center')
            ->where($qb->expr()->in('center', ':centers'))
            ->setParameter(':centers', $centers);
    }

    public function requiredRole()
    {
        return new Role(ActivityStatsVoter::STATS);
    }

    public function supportsModifiers()
    {
        return ['person', 'activity'];
    }
}
