<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\ActivityBundle\Export\Filter;

use Chill\ActivityBundle\Entity\ActivityType;
use Chill\ActivityBundle\Repository\ActivityTypeRepository;
use Chill\ActivityBundle\Security\Authorization\ActivityStatsVoter;
use Chill\MainBundle\Export\ExportElementValidatedInterface;
use Chill\MainBundle\Export\FilterInterface;
use Chill\MainBundle\Templating\TranslatableStringHelperInterface;
use Doctrine\ORM\Query\Expr;
use Doctrine\ORM\Query\Expr\Join;
use Doctrine\ORM\QueryBuilder;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Security\Core\Role\Role;
use Symfony\Component\Validator\Context\ExecutionContextInterface;

use function count;

class ActivityTypeFilter implements ExportElementValidatedInterface, FilterInterface
{
    protected ActivityTypeRepository $activityTypeRepository;

    protected TranslatableStringHelperInterface $translatableStringHelper;

    public function __construct(
        TranslatableStringHelperInterface $translatableStringHelper,
        ActivityTypeRepository $activityTypeRepository
    ) {
        $this->translatableStringHelper = $translatableStringHelper;
        $this->activityTypeRepository = $activityTypeRepository;
    }

    public function addRole()
    {
        return new Role(ActivityStatsVoter::STATS);
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        $where = $qb->getDQLPart('where');
        $clause = $qb->expr()->in('activity.type', ':selected_activity_types');

        if ($where instanceof Expr\Andx) {
            $where->add($clause);
        } else {
            $where = $qb->expr()->andX($clause);
        }

        $qb->add('where', $where);
        $qb->setParameter('selected_activity_types', $data['types']);
    }

    public function applyOn()
    {
        return 'activity';
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder->add(
            'types',
            EntityType::class,
            [
                'class' => ActivityType::class,
                'choice_label' => fn (ActivityType $type) => $this->translatableStringHelper->localize($type->getName()),
                'multiple' => true,
                'expanded' => false,
            ]
        );
    }

    public function describeAction($data, $format = 'string')
    {
        // collect all the reasons'name used in this filter in one array
        $reasonsNames = array_map(
            fn (ActivityType $t): string => '"' . $this->translatableStringHelper->localize($t->getName()) . '"',
            $this->activityTypeRepository->findBy(['id' => $data['types']->toArray()])
        );

        return [
            'Filtered by activity type: only %list%',
            [
                '%list%' => implode(', ', $reasonsNames),
            ],
        ];
    }

    public function getTitle()
    {
        return 'Filter by activity type';
    }

    public function validateForm($data, ExecutionContextInterface $context)
    {
        if (null === $data['types'] || count($data['types']) === 0) {
            $context
                ->buildViolation('At least one type must be chosen')
                ->addViolation();
        }
    }

    /**
     * Check if a join between Activity and Reason is already defined.
     *
     * @param Join[] $joins
     * @param mixed $alias
     *
     * @return bool
     */
    private function checkJoinAlreadyDefined(array $joins, $alias)
    {
        foreach ($joins as $join) {
            if ($join->getAlias() === $alias) {
                return true;
            }
        }

        return false;
    }
}
