<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\ActivityBundle\Export\Filter;

use Chill\ActivityBundle\Entity\ActivityReason;
use Chill\ActivityBundle\Repository\ActivityReasonRepository;
use Chill\MainBundle\Export\ExportElementValidatedInterface;
use Chill\MainBundle\Export\FilterInterface;
use Chill\MainBundle\Form\Type\Export\FilterType;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Chill\MainBundle\Templating\TranslatableStringHelperInterface;
use Chill\PersonBundle\Export\Declarations;
use DateTime;
use Doctrine\ORM\Query\Expr;
use Doctrine\ORM\QueryBuilder;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\Extension\Core\Type\DateType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Form\FormError;
use Symfony\Component\Form\FormEvent;
use Symfony\Component\Form\FormEvents;
use Symfony\Component\Form\FormInterface;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use Symfony\Contracts\Translation\TranslatorInterface;

use function count;

class PersonHavingActivityBetweenDateFilter implements ExportElementValidatedInterface, FilterInterface
{
    protected ActivityReasonRepository $activityReasonRepository;

    protected TranslatableStringHelperInterface $translatableStringHelper;

    protected TranslatorInterface $translator;

    public function __construct(
        TranslatableStringHelper $translatableStringHelper,
        ActivityReasonRepository $activityReasonRepository,
        TranslatorInterface $translator
    ) {
        $this->translatableStringHelper = $translatableStringHelper;
        $this->activityReasonRepository = $activityReasonRepository;
        $this->translator = $translator;
    }

    public function addRole()
    {
        return null;
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        // create a query for activity
        $sqb = $qb->getEntityManager()->createQueryBuilder();
        $sqb->select('person_person_having_activity.id')
            ->from('ChillActivityBundle:Activity', 'activity_person_having_activity')
            ->join('activity_person_having_activity.person', 'person_person_having_activity');

        // add clause between date
        $sqb->where('activity_person_having_activity.date BETWEEN '
            . ':person_having_activity_between_date_from'
            . ' AND '
            . ':person_having_activity_between_date_to');

        // add clause activity reason
        $sqb->join('activity_person_having_activity.reasons', 'reasons_person_having_activity');

        $sqb->andWhere(
            $sqb->expr()->in(
                'reasons_person_having_activity',
                ':person_having_activity_reasons'
            )
        );

        $where = $qb->getDQLPart('where');
        $clause = $qb->expr()->in('person.id', $sqb->getDQL());

        if ($where instanceof Expr\Andx) {
            $where->add($clause);
        } else {
            $where = $qb->expr()->andX($clause);
        }

        $qb->add('where', $where);
        $qb->setParameter(
            'person_having_activity_between_date_from',
            $data['date_from']
        );
        $qb->setParameter(
            'person_having_activity_between_date_to',
            $data['date_to']
        );
        $qb->setParameter('person_having_activity_reasons', $data['reasons']);
    }

    public function applyOn()
    {
        return Declarations::PERSON_IMPLIED_IN;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder->add('date_from', DateType::class, [
            'label' => 'Implied in an activity after this date',
            'data' => new DateTime(),
            'attr' => ['class' => 'datepicker'],
            'widget' => 'single_text',
            'format' => 'dd-MM-yyyy',
        ]);

        $builder->add('date_to', DateType::class, [
            'label' => 'Implied in an activity before this date',
            'data' => new DateTime(),
            'attr' => ['class' => 'datepicker'],
            'widget' => 'single_text',
            'format' => 'dd-MM-yyyy',
        ]);

        $builder->add('reasons', EntityType::class, [
            'class' => ActivityReason::class,
            'choice_label' => fn (ActivityReason $reason): ?string => $this->translatableStringHelper->localize($reason->getName()),
            'group_by' => fn (ActivityReason $reason): ?string => $this->translatableStringHelper->localize($reason->getCategory()->getName()),
            'data' => $this->activityReasonRepository->findAll(),
            'multiple' => true,
            'expanded' => false,
            'label' => 'Activity reasons for those activities',
        ]);

        $builder->addEventListener(FormEvents::POST_SUBMIT, function (FormEvent $event) {
            /** @var FormInterface $filterForm */
            $filterForm = $event->getForm()->getParent();
            $enabled = $filterForm->get(FilterType::ENABLED_FIELD)->getData();

            if (true === $enabled) {
                // if the filter is enabled, add some validation
                $form = $event->getForm();
                $date_from = $form->get('date_from')->getData();
                $date_to = $form->get('date_to')->getData();

                // check that fields are not empty
                if (null === $date_from) {
                    $form->get('date_from')->addError(new FormError(
                        $this->translator->trans('This field '
                        . 'should not be empty')
                    ));
                }

                if (null === $date_to) {
                    $form->get('date_to')->addError(new FormError(
                        $this->translator->trans('This field '
                        . 'should not be empty')
                    ));
                }

                // check that date_from is before date_to
                if (
                    (null !== $date_from && null !== $date_to)
                    && $date_from >= $date_to
                ) {
                    $form->get('date_to')->addError(new FormError(
                        $this->translator->trans('This date '
                        . 'should be after the date given in "Implied in an '
                        . 'activity after this date" field')
                    ));
                }
            }
        });
    }

    public function describeAction($data, $format = 'string')
    {
        return [
            'Filtered by person having an activity between %date_from% and '
            . '%date_to% with reasons %reasons_name%',
            [
                '%date_from%' => $data['date_from']->format('d-m-Y'),
                '%date_to%' => $data['date_to']->format('d-m-Y'),
                '%reasons_name%' => implode(
                    ', ',
                    array_map(
                        fn (ActivityReason $r): string => '"' . $this->translatableStringHelper->localize($r->getName()) . '"',
                        $data['reasons']
                    )
                ),
            ],
        ];
    }

    public function getTitle()
    {
        return 'Filtered by person having an activity in a period';
    }

    public function validateForm($data, ExecutionContextInterface $context)
    {
        if (null === $data['reasons'] || count($data['reasons']) === 0) {
            $context->buildViolation('At least one reason must be chosen')
                ->addViolation();
        }
    }
}
