<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\ActivityBundle\Validator\Constraints;

use Chill\ActivityBundle\Entity\Activity;
use Chill\PersonBundle\Entity\SocialWork\SocialIssue;
use Symfony\Component\Validator\Constraint;
use Symfony\Component\Validator\ConstraintValidator;
use Symfony\Component\Validator\Exception\UnexpectedTypeException;
use Symfony\Component\Validator\Exception\UnexpectedValueException;

use function array_merge;
use function count;

class ActivityValidityValidator extends ConstraintValidator
{
    public function validate($activity, Constraint $constraint)
    {
        if (!$constraint instanceof ActivityValidity) {
            throw new UnexpectedTypeException($constraint, ActivityValidity::class);
        }

        if (!$activity instanceof Activity) {
            throw new UnexpectedValueException($activity, Activity::class);
        }

        if ($activity->getActivityType()->getPersonsVisible() === 2 && count($activity->getPersons()) === 0) {
            $this->context
                ->buildViolation($constraint->noPersonsMessage)
                ->addViolation();
        }

        if ($activity->getActivityType()->getUsersVisible() === 2 && count($activity->getUsers()) === 0) {
            $this->context
                ->buildViolation($constraint->noUsersMessage)
                ->addViolation();
        }

        if ($activity->getActivityType()->getThirdPartiesVisible() === 2 && count($activity->getThirdParties()) === 0) {
            $this->context
                ->buildViolation($constraint->noThirdPartiesMessage)
                ->addViolation();
        }

        if ($activity->getActivityType()->getUserVisible() === 2 && null === $activity->getUser()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('user'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getDateVisible() === 2 && null === $activity->getDate()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('date'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getLocationVisible() === 2 && null === $activity->getLocation()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('location'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getDurationTimeVisible() === 2 && null === $activity->getDurationTime()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('duration time'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getTravelTimeVisible() === 2 && null === $activity->getTravelTime()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('travel time'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getAttendeeVisible() === 2 && null === $activity->getAttendee()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('attendee'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getReasonsVisible() === 2 && null === $activity->getReasons()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('reasons'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getCommentVisible() === 2 && null === $activity->getComment()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('comment'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getSentReceivedVisible() === 2 && null === $activity->getSentReceived()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('sent/received'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getDocumentsVisible() === 2 && null === $activity->getDocuments()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('document'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getEmergencyVisible() === 2 && null === $activity->getEmergency()) {
            $this->context
                ->buildViolation($constraint->makeIsRequiredMessage('emergency'))
                ->addViolation();
        }

        if ($activity->getActivityType()->getSocialIssuesVisible() === 2 && $activity->getSocialIssues()->count() === 0) {
            $this->context
                ->buildViolation($constraint->socialIssuesMessage)
                ->addViolation();
        }

        if ($activity->getActivityType()->getSocialActionsVisible() === 2 && $activity->getSocialActions()->count() === 0) {
            // check if a social action may be added
            $actions = [];

            foreach ($activity->getSocialIssues() as $socialIssue) {
                /** @var SocialIssue $socialIssue */
                $actions = array_merge($actions, $socialIssue->getRecursiveSocialActions()->toArray());
            }

            if (0 < count($actions)) {
                $this->context
                    ->buildViolation($constraint->socialActionsMessage)
                    ->addViolation();
            }
        }
    }
}
