<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\BudgetBundle\Form;

use Chill\BudgetBundle\Config\ConfigRepository;
use Chill\BudgetBundle\Entity\Charge;
use Chill\MainBundle\Form\Type\ChillDateType;
use Chill\MainBundle\Templating\TranslatableStringHelperInterface;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\Extension\Core\Type\MoneyType;
use Symfony\Component\Form\Extension\Core\Type\TextareaType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\OptionsResolver\OptionsResolver;

use function array_flip;
use function asort;

class ChargeType extends AbstractType
{
    protected ConfigRepository $configRepository;

    protected TranslatableStringHelperInterface $translatableStringHelper;

    public function __construct(
        ConfigRepository $configRepository,
        TranslatableStringHelperInterface $translatableStringHelper
    ) {
        $this->configRepository = $configRepository;
        $this->translatableStringHelper = $translatableStringHelper;
    }

    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $builder
            ->add('type', ChoiceType::class, [
                'choices' => $this->getTypes(),
                'placeholder' => 'Choose a charge type',
                'attr' => ['class' => ' select2 '],
            ])
            ->add('amount', MoneyType::class)
            ->add('comment', TextareaType::class, [
                'required' => false,
            ]);

        if ($options['show_start_date']) {
            $builder->add('startDate', ChillDateType::class, [
                'label' => 'Start of validity period',
            ]);
        }

        if ($options['show_end_date']) {
            $builder->add('endDate', ChillDateType::class, [
                'required' => false,
                'label' => 'End of validity period',
            ]);
        }

        if ($options['show_help']) {
            $builder->add('help', ChoiceType::class, [
                'choices' => [
                    'charge.help.running' => Charge::HELP_ASKED,
                    'charge.help.no' => Charge::HELP_NO,
                    'charge.help.yes' => Charge::HELP_YES,
                    'charge.help.not-concerned' => Charge::HELP_NOT_RELEVANT,
                ],
                'placeholder' => 'Choose a status',
                'required' => false,
                'label' => 'Help to pay charges',
            ]);
        }
    }

    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver->setDefaults([
            'data_class' => Charge::class,
            'show_start_date' => true,
            'show_end_date' => true,
            'show_help' => true,
        ]);

        $resolver
            ->setAllowedTypes('show_start_date', 'boolean')
            ->setAllowedTypes('show_end_date', 'boolean')
            ->setAllowedTypes('show_help', 'boolean');
    }

    public function getBlockPrefix()
    {
        return 'chill_budgetbundle_charge';
    }

    private function getTypes()
    {
        $charges = $this->configRepository
            ->getChargesLabels();

        // rewrite labels to filter in language
        foreach ($charges as $key => $labels) {
            $charges[$key] = $this->translatableStringHelper->localize($labels);
        }

        asort($charges);

        return array_flip($charges);
    }
}
