<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\CalendarBundle\Form;

use Chill\CalendarBundle\Entity\Calendar;
use Chill\CalendarBundle\Entity\CalendarRange;
use Chill\CalendarBundle\Entity\CancelReason;
use Chill\CalendarBundle\Entity\Invite;
use Chill\MainBundle\Entity\Location;
use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Form\Type\CommentType;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Chill\PersonBundle\Entity\Person;
use Chill\ThirdPartyBundle\Entity\ThirdParty;
use DateTimeImmutable;
use Doctrine\Persistence\ObjectManager;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\CallbackTransformer;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\Extension\Core\Type\HiddenType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\OptionsResolver\OptionsResolver;

class CalendarType extends AbstractType
{
    protected ObjectManager $om;

    protected TranslatableStringHelper $translatableStringHelper;

    public function __construct(
        TranslatableStringHelper $translatableStringHelper,
        ObjectManager $om
    ) {
        $this->translatableStringHelper = $translatableStringHelper;
        $this->om = $om;
    }

    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $builder
            ->add('comment', CommentType::class, [
                'required' => false,
            ])
            // ->add('cancelReason', EntityType::class, [
            //     'required' => false,
            //     'class' => CancelReason::class,
            //     'choice_label' => function (CancelReason $entity) {
            //         return $entity->getCanceledBy();
            //     },
            // ])
            ->add('sendSMS', ChoiceType::class, [
                'required' => false,
                'choices' => [
                    'Oui' => true,
                    'Non' => false,
                ],
                'expanded' => true,
            ]);

        $builder->add('mainUser', HiddenType::class);
        $builder->get('mainUser')
            ->addModelTransformer(new CallbackTransformer(
                static function (?User $user): int {
                    if (null !== $user) {
                        $res = $user->getId();
                    } else {
                        $res = -1; //TODO cannot be null in any ways...
                    }

                    return $res;
                },
                function (?int $userId): User {
                    return $this->om->getRepository(user::class)->findOneBy(['id' => (int) $userId]);
                }
            ));

        $builder->add('startDate', HiddenType::class);
        $builder->get('startDate')
            ->addModelTransformer(new CallbackTransformer(
                static function (?DateTimeImmutable $dateTimeImmutable): string {
                    if (null !== $dateTimeImmutable) {
                        $res = date_format($dateTimeImmutable, 'Y-m-d H:i:s');
                    } else {
                        $res = '';
                    }

                    return $res;
                },
                static function (?string $dateAsString): DateTimeImmutable {
                    return new DateTimeImmutable($dateAsString);
                }
            ));

        $builder->add('endDate', HiddenType::class);
        $builder->get('endDate')
            ->addModelTransformer(new CallbackTransformer(
                static function (?DateTimeImmutable $dateTimeImmutable): string {
                    if (null !== $dateTimeImmutable) {
                        $res = date_format($dateTimeImmutable, 'Y-m-d H:i:s');
                    } else {
                        $res = '';
                    }

                    return $res;
                },
                static function (?string $dateAsString): DateTimeImmutable {
                    return new DateTimeImmutable($dateAsString);
                }
            ));

        $builder->add('persons', HiddenType::class);
        $builder->get('persons')
            ->addModelTransformer(new CallbackTransformer(
                static function (iterable $personsAsIterable): string {
                    $personIds = [];

                    foreach ($personsAsIterable as $value) {
                        $personIds[] = $value->getId();
                    }

                    return implode(',', $personIds);
                },
                function (?string $personsAsString): array {
                    if (null === $personsAsString) {
                        return [];
                    }
                    return array_map(
                        fn (string $id): ?Person => $this->om->getRepository(Person::class)->findOneBy(['id' => (int) $id]),
                        explode(',', $personsAsString)
                    );
                }
            ));

        $builder->add('professionals', HiddenType::class);
        $builder->get('professionals')
            ->addModelTransformer(new CallbackTransformer(
                static function (iterable $thirdpartyAsIterable): string {
                    $thirdpartyIds = [];

                    foreach ($thirdpartyAsIterable as $value) {
                        $thirdpartyIds[] = $value->getId();
                    }

                    return implode(',', $thirdpartyIds);
                },
                function (?string $thirdpartyAsString): array {
                    if (null === $thirdpartyAsString) {
                        return [];
                    }
                    return array_map(
                        fn (string $id): ?ThirdParty => $this->om->getRepository(ThirdParty::class)->findOneBy(['id' => (int) $id]),
                        explode(',', $thirdpartyAsString)
                    );
                }
            ));

        $builder->add('calendarRange', HiddenType::class);
        $builder->get('calendarRange')
            ->addModelTransformer(new CallbackTransformer(
                static function (?CalendarRange $calendarRange): ?int {
                    if (null !== $calendarRange) {
                        $res = $calendarRange->getId();
                    } else {
                        $res = -1;
                    }

                    return $res;
                },
                function (?string $calendarRangeId): ?CalendarRange {
                    if (null !== $calendarRangeId) {
                        $res = $this->om->getRepository(CalendarRange::class)->findOneBy(['id' => (int) $calendarRangeId]);
                    } else {
                        $res = null;
                    }

                    return $res;
                }
            ));

        $builder->add('location', HiddenType::class)
            ->get('location')
            ->addModelTransformer(new CallbackTransformer(
                static function (?Location $location): string {
                    if (null === $location) {
                        return '';
                    }

                    return $location->getId();
                },
                function (?string $id): ?Location {
                    return $this->om->getRepository(Location::class)->findOneBy(['id' => (int) $id]);
                }
            ));

        $builder->add('invites', HiddenType::class);
        $builder->get('invites')
            ->addModelTransformer(new CallbackTransformer(
                static function (iterable $usersAsIterable): string {
                    $userIds = [];

                    foreach ($usersAsIterable as $value) {
                        $userIds[] = $value->getId();
                    }

                    return implode(',', $userIds);
                },
                function (?string $usersAsString): array {
                    return array_map(
                        fn (string $id): ?Invite => $this->om->getRepository(Invite::class)->findOneBy(['id' => (int) $id]),
                        explode(',', $usersAsString)
                    );
                }
            ));
    }

    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver->setDefaults([
            'data_class' => Calendar::class,
        ]);

        $resolver
            ->setRequired(['accompanyingPeriod'])
            ->setAllowedTypes('accompanyingPeriod', [\Chill\PersonBundle\Entity\AccompanyingPeriod::class, 'null']);
    }

    public function getBlockPrefix()
    {
        // as the js share some hardcoded items from activity, we have to rewrite block prefix
        return 'chill_activitybundle_activity';
    }
}
