<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\DocGeneratorBundle\tests\Serializer\Normalizer;

use Chill\MainBundle\Entity\Scope;
use Chill\MainBundle\Entity\User;
use Symfony\Bundle\FrameworkBundle\Test\KernelTestCase;
use Symfony\Component\Serializer\Annotation as Serializer;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * @internal
 * @coversNothing
 */
final class DocGenObjectNormalizerTest extends KernelTestCase
{
    private NormalizerInterface $normalizer;

    protected function setUp(): void
    {
        parent::setUp();
        self::bootKernel();

        $this->normalizer = self::$container->get(NormalizerInterface::class);
    }

    public function testChangeContextOnAttribute()
    {
        $object = new TestableParentClass();
        $actual = $this->normalizer->normalize(
            $object,
            'docgen',
            ['groups' => 'docgen:read']
        );

        $this->assertIsArray($actual);
        $this->assertArrayHasKey('child', $actual);
        $this->assertIsArray($actual['child']);
        $this->assertArrayHasKey('foo', $actual['child']);
        $this->assertEquals('bar', $actual['child']['foo']);
        $this->assertArrayNotHasKey('baz', $actual['child']);

        // test with child = null
        $object->child = null;
        $actual = $this->normalizer->normalize(
            $object,
            'docgen',
            ['groups' => 'docgen:read']
        );
        $this->assertIsArray($actual);
        $this->assertArrayHasKey('child', $actual);
        $this->assertIsArray($actual['child']);
        $this->assertArrayHasKey('foo', $actual['child']);
        $this->assertEquals('', $actual['child']['foo']);
        $this->assertArrayNotHasKey('baz', $actual['child']);

        $actual = $this->normalizer->normalize(
            null,
            'docgen',
            ['groups' => 'docgen:read', 'docgen:expects' => TestableParentClass::class],
        );
        $this->assertIsArray($actual);
        $this->assertArrayHasKey('child', $actual);
        $this->assertIsArray($actual['child']);
        $this->assertArrayHasKey('foo', $actual['child']);
        $this->assertEquals('', $actual['child']['foo']);
        $this->assertArrayNotHasKey('baz', $actual['child']);
    }

    public function testNormalizationBasic()
    {
        $scope = new Scope();
        $scope->setName(['fr' => 'scope']);

        $normalized = $this->normalizer->normalize($scope, 'docgen', [AbstractNormalizer::GROUPS => ['docgen:read'], 'docgen:expects' => Scope::class]);
        $expected = [
            'id' => null,
            'name' => 'scope',
            'type' => 'scope',
            'isNull' => false,
        ];

        $this->assertEquals($expected, $normalized, 'test normalization fo a scope');
    }

    public function testNormalizeNull()
    {
        $actual = $this->normalizer->normalize(null, 'docgen', [AbstractNormalizer::GROUPS => ['docgen:read'], 'docgen:expects' => Scope::class]);
        $expected = [
            'id' => '',
            'name' => '',
            'type' => 'scope',
            'isNull' => true,
        ];

        $this->assertEquals($expected, $actual, 'test normalization for a null scope');
    }

    public function testNormalizeNullObjectWithObjectEmbedded()
    {
        $this->markTestIncomplete('test to implement');
        $normalized = $this->normalizer->normalize(null, 'docgen', [
            AbstractNormalizer::GROUPS => ['docgen:read'],
            'docgen:expects' => User::class,
        ]);

        $expected = [
            'label' => '',
            'email' => '',
            'main_center' => [
                'name' => '',
            ],
        ];

        $this->assertEquals($expected, $normalized, 'test normalization for a null user');
    }

    public function testNormalizeWithNullValueEmbedded()
    {
        $this->markTestIncomplete('test to write');
        $user = new User();
        $user->setUsername('User Test');

        $normalized = $this->normalizer->normalize($user, 'docgen', [AbstractNormalizer::GROUPS => ['docgen:read'], 'docgen:expects' => User::class]);
        $expected = [
            'label' => 'User Test',
            'email' => '',
            'mainCenter' => [
                'name' => '',
            ],
        ];

        $this->assertEquals($expected, $normalized, 'test normalization fo an user with null center');
    }
}

class TestableParentClass
{
    /**
     * @Serializer\Groups("docgen:read")
     * @Serializer\Context(normalizationContext={"groups": "docgen:read:foo"}, groups={"docgen:read"})
     */
    public ?TestableChildClass $child;

    public function __construct()
    {
        $this->child = new TestableChildClass();
    }
}

class TestableChildClass
{
    /**
     * @Serializer\Groups("docgen:read")
     */
    public string $baz = 'bloup';

    /**
     * @Serializer\Groups("docgen:read:foo")
     */
    public string $foo = 'bar';
}
