<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\EventBundle\Form\Type;

use Chill\EventBundle\Entity\EventType;
use Chill\EventBundle\Entity\Status;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Doctrine\ORM\EntityRepository;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Form\FormEvent;
use Symfony\Component\Form\FormEvents;
use Symfony\Component\OptionsResolver\OptionsResolver;
use Symfony\Contracts\Translation\TranslatorInterface;

/**
 * Allow to pick amongst type.
 *
 * parameters :
 *
 * - event_type : restricts to a certain event type. Default null (= all event types)
 * - active_only: restricts to active type only. Default true
 */
class PickStatusType extends AbstractType
{
    /**
     * @var EntityRepository
     */
    protected $statusRepository;

    /**
     * @var TranslatableStringHelper
     */
    protected $translatableStringHelper;

    /**
     * @var TranslatorInterface
     */
    protected $translator;

    public function __construct(
        TranslatableStringHelper $translatableStringHelper,
        TranslatorInterface $translator,
        EntityRepository $statusRepository
    ) {
        $this->translatableStringHelper = $translatableStringHelper;
        $this->translator = $translator;
        $this->statusRepository = $statusRepository;
    }

    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $qb = $options['query_builder'];

        if ($options['event_type'] instanceof EventType) {
            $options['query_builder']->where($qb->expr()->eq('r.type', ':event_type'))
                ->setParameter('event_type', $options['event_type']);
        }

        if (true === $options['active_only']) {
            $options['query_builder']->andWhere($qb->expr()->eq('r.active', ':active'))
                ->setParameter('active', true);
        }

        if (null === $options['group_by'] && null === $options['event_type']) {
            $builder->addEventListener(
                FormEvents::PRE_SET_DATA,
                function (FormEvent $event) {
                    $form = $event->getForm();
                    $name = $form->getName();
                    $config = $form->getConfig();
                    $type = $config->getType()->getName();
                    $options = $config->getOptions();
                    $form->getParent()->add($name, $type, array_replace($options, [
                        'group_by' => function (Status $s) {
                            return $this->translatableStringHelper->localize($s->getType()->getName());
                        },
                    ]));
                }
            );
        }
    }

    public function configureOptions(OptionsResolver $resolver)
    {
        // create copy for use in Closure
        $translatableStringHelper = $this->translatableStringHelper;
        $translator = $this->translator;

        $resolver
              // add option "event_type"
            ->setDefined('event_type')
            ->setAllowedTypes('event_type', ['null', EventType::class])
            ->setDefault('event_type', null)
            // add option allow unactive
            ->setDefault('active_only', true)
            ->setAllowedTypes('active_only', ['boolean']);

        $qb = $this->statusRepository->createQueryBuilder('r');

        $resolver->setDefaults([
            'class' => Status::class,
            'query_builder' => $qb,
            'group_by' => null,
            'choice_attr' => static function (Status $s) {
                return [
                    'data-event-type' => $s->getType()->getId(),
                    'data-link-category' => $s->getType()->getId(),
                ];
            },
            'choice_label' => static function (Status $s) use ($translatableStringHelper, $translator) {
                return $translatableStringHelper->localize($s->getName()) .
                          ($s->getActive() === true ? '' :
                                ' (' . $translator->trans('unactive') . ')');
            },
        ]);
    }

    public function getParent()
    {
        return EntityType::class;
    }
}
