<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\MainBundle\Command;

use Chill\MainBundle\Entity\Language;
use Doctrine\ORM\EntityManager;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Intl\Intl;

use function in_array;

/*
 * Load or update the languages entities command
 */
class LoadAndUpdateLanguagesCommand extends Command
{
    public const INCLUDE_ANCIENT = 'include_ancient';

    public const INCLUDE_REGIONAL_VERSION = 'include_regional';

    // Array of ancien languages (to exclude)
    private $ancientToExclude = ['ang', 'egy', 'fro', 'goh', 'grc', 'la', 'non', 'peo', 'pro', 'sga',
        'dum', 'enm', 'frm', 'gmh', 'mga', 'akk', 'phn', 'zxx', 'got', 'und', ];

    private $availableLanguages;

    /**
     * @var EntityManager
     */
    private $entityManager;

    // The regional version of language are language with _ in the code
    // This array contains regional code to not exclude
    private $regionalVersionToInclude = ['ro_MD'];

    /**
     * LoadCountriesCommand constructor.
     *
     * @param $availableLanguages
     */
    public function __construct(EntityManager $entityManager, $availableLanguages)
    {
        $this->entityManager = $entityManager;
        $this->availableLanguages = $availableLanguages;
        parent::__construct();
    }

    /**
     * (non-PHPdoc).
     *
     * @see \Symfony\Component\Console\Command\Command::configure()
     */
    protected function configure()
    {
        $this
            ->setName('chill:main:languages:populate')
            ->setDescription('Load or update languages in db. This command does not delete existing ' .
                'languages, but will update names according to available languages')
            ->addOption(
                self::INCLUDE_REGIONAL_VERSION,
                null,
                InputOption::VALUE_NONE,
                'Include the regional languages. The regional languages are languages with code containing _ excepted '
                        . implode(',', $this->regionalVersionToInclude) . '.'
            )
            ->addOption(
                self::INCLUDE_ANCIENT,
                null,
                InputOption::VALUE_NONE,
                'Include the ancient languages that are languages with code '
                        . implode(', ', $this->ancientToExclude) . '.'
            );
    }

    /**
     * (non-PHPdoc).
     *
     * @see \Symfony\Component\Console\Command\Command::execute()
     */
    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $em = $this->entityManager;
        $chillAvailableLanguages = $this->availableLanguages;
        $languageBundle = Intl::getLanguageBundle();
        $languages = [];

        foreach ($chillAvailableLanguages as $avLang) {
            $languages[$avLang] = $languageBundle->getLanguageNames($avLang);
        }

        $languageCodes = array_keys($languages[$chillAvailableLanguages[0]]);

        foreach ($languageCodes as $code) {
            $excludeCode = (
                (
                    !$input->getOption(self::INCLUDE_REGIONAL_VERSION)
                    && strpos($code, '_')
                    && !in_array($code, $this->regionalVersionToInclude, true)
                ) || (
                    !$input->getOption(self::INCLUDE_ANCIENT)
                    && in_array($code, $this->ancientToExclude, true)
                )
            );

            $langageDB = $em->getRepository(Language::class)->find($code);

            if (!$excludeCode) {
                if (!$langageDB) {
                    $langageDB = new \Chill\MainBundle\Entity\Language();
                    $langageDB->setId($code);
                    $em->persist($langageDB);
                }

                $avLangNames = [];

                foreach ($chillAvailableLanguages as $avLang) {
                    $avLangNames[$avLang] = $languages[$avLang][$code];
                }

                $langageDB->setName($avLangNames);
            } else {
                if ($langageDB) {
                    $em->remove($langageDB);
                }
                echo 'Code excluded : ' . $code . ' - ' . $languageBundle->getLanguageName($code) . "\n";
            }
        }

        $em->flush();
    }
}
