<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\MainBundle\Form\Type;

use Chill\MainBundle\Entity\Scope;
use Chill\MainBundle\Security\RoleProvider;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\OptionsResolver\OptionsResolver;

use function in_array;

/**
 * Form to Edit/create a role scope. If the role scope does not
 * exists in the database, he is generated.
 */
class ComposedRoleScopeType extends AbstractType
{
    /**
     * @var RoleProvider
     */
    private $roleProvider;

    /**
     * @var string[]
     */
    private $roles = [];

    /**
     * @var string[]
     */
    private $rolesWithoutScope = [];

    /**
     * @var TranslatableStringHelper
     */
    private $translatableStringHelper;

    public function __construct(
        TranslatableStringHelper $translatableStringHelper,
        RoleProvider $roleProvider
    ) {
        $this->roles = $roleProvider->getRoles();
        $this->rolesWithoutScope = $roleProvider->getRolesWithoutScopes();
        $this->translatableStringHelper = $translatableStringHelper;
        $this->roleProvider = $roleProvider;
    }

    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        // store values used in internal function
        $translatableStringHelper = $this->translatableStringHelper;
        $rolesWithoutScopes = $this->rolesWithoutScope;

        //build roles
        $values = [];

        foreach ($this->roles as $role) {
            $values[$role] = $role;
        }

        $builder
            ->add('role', ChoiceType::class, [
                'choices' => array_combine(array_values($values), array_keys($values)),
                'placeholder' => 'Choose amongst roles',
                'choice_attr' => static function ($role) use ($rolesWithoutScopes) {
                    if (in_array($role, $rolesWithoutScopes, true)) {
                        return ['data-has-scope' => '0'];
                    }

                    return ['data-has-scope' => '1'];
                },
                'group_by' => function ($role, $key, $index) {
                    return $this->roleProvider->getRoleTitle($role);
                },
            ])
            ->add('scope', EntityType::class, [
                'class' => 'ChillMainBundle:Scope',
                'choice_label' => static function (Scope $scope) use ($translatableStringHelper) {
                    return $translatableStringHelper->localize($scope->getName());
                },
                'required' => false,
                'data' => null,
            ]);
    }

    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver->setDefault('data_class', 'Chill\MainBundle\Entity\RoleScope');
    }
}
