<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\MainBundle\Form\Type\Export;

use Chill\MainBundle\Export\ExportManager;
use Chill\MainBundle\Validator\Constraints\Export\ExportElementConstraint;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\Extension\Core\Type\FormType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\OptionsResolver\OptionsResolver;

class ExportType extends AbstractType
{
    public const AGGREGATOR_KEY = 'aggregators';

    public const EXPORT_KEY = 'export';

    public const FILTER_KEY = 'filters';

    public const PICK_FORMATTER_KEY = 'pick_formatter';

    /**
     * @var ExportManager
     */
    protected $exportManager;

    public function __construct(ExportManager $exportManager)
    {
        $this->exportManager = $exportManager;
    }

    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $export = $this->exportManager->getExport($options['export_alias']);

        $exportOptions = [
            'compound' => true,
            'constraints' => [
            ],
        ];

        // add a contraint if required by export
        $exportBuilder = $builder->create(self::EXPORT_KEY/*, FormType::class, $exportOptions*/);

        $export->buildForm($exportBuilder);

        $builder->add($exportBuilder, null, $exportOptions);

        if ($export instanceof \Chill\MainBundle\Export\ExportInterface) {
            //add filters
            $filters = $this->exportManager->getFiltersApplyingOn($export, $options['picked_centers']);
            $filterBuilder = $builder->create(self::FILTER_KEY, FormType::class, ['compound' => true]);

            foreach ($filters as $alias => $filter) {
                $filterBuilder->add($alias, FilterType::class, [
                    'filter_alias' => $alias,
                    'export_manager' => $this->exportManager,
                    'label' => $filter->getTitle(),
                    'constraints' => [
                        new ExportElementConstraint(['element' => $filter]),
                    ],
                ]);
            }

            $builder->add($filterBuilder);

            //add aggregators
            $aggregators = $this->exportManager
                ->getAggregatorsApplyingOn($export, $options['picked_centers']);
            $aggregatorBuilder = $builder->create(
                self::AGGREGATOR_KEY,
                FormType::class,
                ['compound' => true]
            );

            foreach ($aggregators as $alias => $aggregator) {
                $aggregatorBuilder->add($alias, AggregatorType::class, [
                    'aggregator_alias' => $alias,
                    'export_manager' => $this->exportManager,
                    'label' => $aggregator->getTitle(),
                    'constraints' => [
                        new ExportElementConstraint(['element' => $aggregator]),
                    ],
                ]);
            }

            $builder->add($aggregatorBuilder);
        }

        // add export form
        $exportBuilder = $builder->create(self::EXPORT_KEY, FormType::class, ['compound' => true]);
        $this->exportManager->getExport($options['export_alias'])
            ->buildForm($exportBuilder);
        $builder->add($exportBuilder);

        if ($export instanceof \Chill\MainBundle\Export\ExportInterface) {
            $builder->add(self::PICK_FORMATTER_KEY, PickFormatterType::class, [
                'export_alias' => $options['export_alias'],
            ]);
        }
    }

    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver->setRequired(['export_alias', 'picked_centers'])
            ->setAllowedTypes('export_alias', ['string'])
            ->setDefault('compound', true)
            ->setDefault('constraints', [
                //new \Chill\MainBundle\Validator\Constraints\Export\ExportElementConstraint()
            ]);
    }

    public function getParent()
    {
        return FormType::class;
    }
}
