<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\MainBundle\Form\Type\Export;

use Chill\MainBundle\Center\GroupingCenterInterface;
use Chill\MainBundle\Entity\Center;
use Chill\MainBundle\Export\ExportManager;
use Chill\MainBundle\Security\Authorization\AuthorizationHelper;
use Doctrine\ORM\EntityRepository;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\CallbackTransformer;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\OptionsResolver\OptionsResolver;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

use function array_intersect;
use function array_key_exists;
use function array_merge;
use function array_unique;
use function count;
use function in_array;

/**
 * Pick centers amongst available centers for the user.
 */
class PickCenterType extends AbstractType
{
    public const CENTERS_IDENTIFIERS = 'c';

    /**
     * @var AuthorizationHelper
     */
    protected $authorizationHelper;

    /**
     * @var ExportManager
     */
    protected $exportManager;

    /**
     * @var GroupingCenterInterface[]
     */
    protected $groupingCenters = [];

    /**
     * @var \Symfony\Component\Security\Core\User\UserInterface
     */
    protected $user;

    public function __construct(
        TokenStorageInterface $tokenStorage,
        ExportManager $exportManager,
        AuthorizationHelper $authorizationHelper
    ) {
        $this->exportManager = $exportManager;
        $this->user = $tokenStorage->getToken()->getUser();
        $this->authorizationHelper = $authorizationHelper;
    }

    public function addGroupingCenter(GroupingCenterInterface $grouping)
    {
        $this->groupingCenters[md5($grouping->getName())] = $grouping;
    }

    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $export = $this->exportManager->getExport($options['export_alias']);
        $centers = $this->authorizationHelper->getReachableCenters(
            $this->user,
            $export->requiredRole()
        );

        $builder->add(self::CENTERS_IDENTIFIERS, EntityType::class, [
            'class' => 'ChillMainBundle:Center',
            'query_builder' => static function (EntityRepository $er) use ($centers) {
                $qb = $er->createQueryBuilder('c');
                $ids = array_map(
                    static function (Center $el) {
                        return $el->getId();
                    },
                    $centers
                );

                return $qb->where($qb->expr()->in('c.id', $ids));
            },
            'multiple' => true,
            'expanded' => true,
            'choice_label' => static function (Center $c) {
                return $c->getName();
            },
            'data' => count($this->groupingCenters) > 0 ? null : $centers,
        ]);

        if (count($this->groupingCenters) > 0) {
            $groupingBuilder = $builder->create('g', null, [
                'compound' => true,
            ]);

            foreach ($this->groupingCenters as $key => $gc) {
                $choices = $this->buildChoices($centers, $gc);

                if (count($choices) > 0) {
                    $groupingBuilder->add($key, ChoiceType::class, [
                        'choices' => $choices,
                        'multiple' => true,
                        'expanded' => true,
                        'label' => $gc->getName(),
                        'required' => false,
                    ]);
                }
            }

            if ($groupingBuilder->count() > 0) {
                $builder->add($groupingBuilder);
            }
        }

        $builder->addModelTransformer(new CallbackTransformer(
            function ($data) use ($centers) {
                return $this->transform($data, $centers);
            },
            function ($data) use ($centers) {
                return $this->reverseTransform($data, $centers);
            }
        ));
    }

    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver->setRequired('export_alias');
    }

    protected function buildChoices($reachablesCenters, GroupingCenterInterface $gc)
    {
        $result = [];

        foreach ($gc->getGroups() as $group) {
            foreach ($gc->getCentersForGroup($group) as $center) {
                if (in_array($center, $reachablesCenters, true)) {
                    $result[$group] = $group;
                }
            }
        }

        return $result;
    }

    protected function reverseTransform($data, $centers)
    {
        $picked = $data[self::CENTERS_IDENTIFIERS]
            instanceof \Doctrine\Common\Collections\Collection ?
                $data[self::CENTERS_IDENTIFIERS]->toArray()
                :
                $data[self::CENTERS_IDENTIFIERS];

        if (array_key_exists('g', $data)) {
            foreach ($data['g'] as $gcid => $group) {
                $picked =
                    array_merge(
                        array_intersect(
                            $this->groupingCenters[$gcid]->getCentersForGroup($group),
                            $centers
                        ),
                        $picked
                    );
            }
        }

        return array_unique($picked);
    }

    protected function transform($data, $centers)
    {
        return $data;
    }
}
