<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\MainBundle\Form\Type;

use Chill\MainBundle\Entity\Center;
use Chill\MainBundle\Entity\Scope;
use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Form\DataMapper\ScopePickerDataMapper;
use Chill\MainBundle\Repository\ScopeRepository;
use Chill\MainBundle\Security\Authorization\AuthorizationHelperInterface;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\AbstractType;
use Symfony\Component\Form\Extension\Core\Type\HiddenType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Form\FormInterface;
use Symfony\Component\Form\FormView;
use Symfony\Component\OptionsResolver\Options;
use Symfony\Component\OptionsResolver\OptionsResolver;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Security\Core\Role\Role;
use Symfony\Component\Security\Core\Security;

use function array_map;
use function count;

/**
 * Allow to pick amongst available scope for the current
 * user.
 *
 * options :
 *
 * - `center`: the center of the entity
 * - `role`  : the role of the user
 */
class ScopePickerType extends AbstractType
{
    protected AuthorizationHelperInterface $authorizationHelper;

    /**
     * @var ScopeRepository
     */
    protected $scopeRepository;

    protected Security $security;

    /**
     * @var TokenStorageInterface
     */
    protected $tokenStorage;

    /**
     * @var TranslatableStringHelper
     */
    protected $translatableStringHelper;

    public function __construct(
        AuthorizationHelperInterface $authorizationHelper,
        TokenStorageInterface $tokenStorage,
        ScopeRepository $scopeRepository,
        Security $security,
        TranslatableStringHelper $translatableStringHelper
    ) {
        $this->authorizationHelper = $authorizationHelper;
        $this->tokenStorage = $tokenStorage;
        $this->scopeRepository = $scopeRepository;
        $this->security = $security;
        $this->translatableStringHelper = $translatableStringHelper;
    }

    public function buildForm(FormBuilderInterface $builder, array $options)
    {
        $items = $this->authorizationHelper->getReachableScopes(
            $this->security->getUser(),
            $options['role'] instanceof Role ? $options['role']->getRole() : $options['role'],
            $options['center']
        );

        if (1 !== count($items)) {
            $builder->add('scope', EntityType::class, [
                'class' => Scope::class,
                'placeholder' => 'Choose the circle',
                'choice_label' => function (Scope $c) {
                    return $this->translatableStringHelper->localize($c->getName());
                },
                'choices' => $items,
            ]);
            $builder->setDataMapper(new ScopePickerDataMapper());
        } else {
            $builder->add('scope', HiddenType::class, [
                'data' => $items[0]->getId(),
            ]);
            $builder->setDataMapper(new ScopePickerDataMapper($items[0]));
        }
    }

    public function buildView(FormView $view, FormInterface $form, array $options)
    {
        $view->vars = array_replace(
            $view->vars,
            [
                'hideLabel' => true,
            ]
        );
    }

    public function configureOptions(OptionsResolver $resolver)
    {
        $resolver
          // create `center` option
            ->setRequired('center')
            ->setAllowedTypes('center', [Center::class, 'array', 'null'])
          // create ``role` option
            ->setRequired('role')
            ->setAllowedTypes('role', ['string', Role::class]);
    }

    /**
     * @param array|Center|Center[] $center
     * @param string $role
     *
     * @return \Doctrine\ORM\QueryBuilder
     */
    protected function buildAccessibleScopeQuery($center, $role)
    {
        $roles = $this->authorizationHelper->getParentRoles($role);
        $roles[] = $role;
        $centers = $center instanceof Center ? [$center] : $center;

        $qb = $this->scopeRepository->createQueryBuilder('s');
        $qb
          // jointure to center
            ->join('s.roleScopes', 'rs')
            ->join('rs.permissionsGroups', 'pg')
            ->join('pg.groupCenters', 'gc')
          // add center constraint
            ->where($qb->expr()->in('IDENTITY(gc.center)', ':centers'))
            ->setParameter('centers', array_map(static fn (Center $c) => $c->getId(), $centers))
          // role constraints
            ->andWhere($qb->expr()->in('rs.role', ':roles'))
            ->setParameter('roles', $roles)
          // user contraint
            ->andWhere(':user MEMBER OF gc.users')
            ->setParameter('user', $this->tokenStorage->getToken()->getUser());

        return $qb;
    }
}
