<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\MainBundle\Security\PasswordRecover;

use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Notification\Mailer;
use DateTimeInterface;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;

use function array_merge;

class RecoverPasswordHelper
{
    public const RECOVER_PASSWORD_ROUTE = 'password_recover';

    /**
     * @var Mailer
     */
    protected $mailer;

    protected $routeParameters;

    /**
     * @var TokenManager
     */
    protected $tokenManager;

    /**
     * @var UrlGeneratorInterface
     */
    protected $urlGenerator;

    public function __construct(
        TokenManager $tokenManager,
        UrlGeneratorInterface $urlGenerator,
        Mailer $mailer,
        array $routeParameters
    ) {
        $this->tokenManager = $tokenManager;
        $this->urlGenerator = $urlGenerator;
        $this->mailer = $mailer;
        $this->routeParameters = $routeParameters;
    }

    /**
     * @param bool $absolute
     * @param array $parameters additional parameters to url
     *
     * @return string
     */
    public function generateUrl(User $user, DateTimeInterface $expiration, $absolute = true, array $parameters = [])
    {
        $context = $this->urlGenerator->getContext();
        $previousHost = $context->getHost();
        $previousScheme = $context->getScheme();

        $context->setHost($this->routeParameters['host']);
        $context->setScheme($this->routeParameters['scheme']);

        $url = $this->urlGenerator->generate(
            self::RECOVER_PASSWORD_ROUTE,
            array_merge(
                $this->tokenManager->generate($user, $expiration),
                $parameters
            ),
            $absolute ? UrlGeneratorInterface::ABSOLUTE_URL : UrlGeneratorInterface::ABSOLUTE_PATH
        );

        // reset the host
        $context->setHost($previousHost);
        $context->setScheme($previousScheme);

        return $url;
    }

    public function sendRecoverEmail(
        User $user,
        DateTimeInterface $expiration,
        $template = '@ChillMain/Password/recover_email.txt.twig',
        array $templateParameters = [],
        $force = false,
        array $additionalUrlParameters = [],
        $emailSubject = 'Recover your password'
    ) {
        $content = $this->mailer->renderContentToUser(
            $user,
            $template,
            array_merge(
                [
                    'user' => $user,
                    'url' => $this->generateUrl($user, $expiration, true, $additionalUrlParameters),
                ],
                $templateParameters
            )
        );

        $this->mailer->sendNotification(
            $user,
            [$emailSubject],
            [
                'text/plain' => $content,
            ],
            null,
            $force
        );
    }
}
