<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\PersonBundle\Controller;

use Chill\MainBundle\Entity\Address;
use Chill\MainBundle\Form\Type\AddressType;
use Chill\PersonBundle\Entity\Person;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\Form\Extension\Core\Type\SubmitType;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Validator\Validator\ValidatorInterface;

use function count;

/**
 * Class PersonAddressController
 * Controller for addresses associated with person.
 */
class PersonAddressController extends AbstractController
{
    /**
     * @var ValidatorInterface
     */
    protected $validator;

    /**
     * PersonAddressController constructor.
     */
    public function __construct(ValidatorInterface $validator)
    {
        $this->validator = $validator;
    }

    public function createAction($person_id, Request $request)
    {
        $person = $this->getDoctrine()->getManager()
            ->getRepository(\Chill\PersonBundle\Entity\Person::class)
            ->find($person_id);

        if (null === $person) {
            throw $this->createNotFoundException("Person with id {$person_id} not"
                    . ' found ');
        }

        $this->denyAccessUnlessGranted(
            'CHILL_PERSON_UPDATE',
            $person,
            'You are not allowed to edit this person.'
        );

        $address = new Address();

        $form = $this->createCreateForm($person, $address);
        $form->handleRequest($request);

        $person->addAddress($address);

        if ($form->isSubmitted() && $form->isValid()) {
            $validatePersonErrors = $this->validatePerson($person);

            if (count($validatePersonErrors) !== 0) {
                foreach ($validatePersonErrors as $error) {
                    $this->addFlash('error', $error->getMessage());
                }
            } elseif ($form->isValid()) {
                $em = $this->getDoctrine()->getManager();
                $em->flush();

                $this->addFlash(
                    'success',
                    $this->get('translator')->trans('The new address was created successfully')
                );

                return $this->redirectToRoute('chill_person_address_list', [
                    'person_id' => $person->getId(),
                ]);
            } else {
                $this->addFlash('error', $this->get('translator')
                    ->trans('Error! Address not created!'));
            }
        }

        return $this->render('ChillPersonBundle:Address:new.html.twig', [
            'person' => $person,
            'form' => $form->createView(),
        ]);
    }

    public function editAction($person_id, $address_id)
    {
        $person = $this->getDoctrine()->getManager()
            ->getRepository(\Chill\PersonBundle\Entity\Person::class)
            ->find($person_id);

        if (null === $person) {
            throw $this->createNotFoundException("Person with id {$person_id} not"
                    . ' found ');
        }

        $this->denyAccessUnlessGranted(
            'CHILL_PERSON_UPDATE',
            $person,
            'You are not allowed to edit this person.'
        );

        $address = $this->findAddressById($person, $address_id);

        $form = $this->createEditForm($person, $address);

        return $this->render('ChillPersonBundle:Address:edit.html.twig', [
            'person' => $person,
            'address' => $address,
            'form' => $form->createView(),
        ]);
    }

    public function listAction($person_id)
    {
        $person = $this->getDoctrine()->getManager()
            ->getRepository(\Chill\PersonBundle\Entity\Person::class)
            ->find($person_id);

        if (null === $person) {
            throw $this->createNotFoundException("Person with id {$person_id} not"
                    . ' found ');
        }

        $this->denyAccessUnlessGranted(
            'CHILL_PERSON_SEE',
            $person,
            'You are not allowed to edit this person.'
        );

        return $this->render('ChillPersonBundle:Address:list.html.twig', [
            'person' => $person,
        ]);
    }

    public function newAction($person_id)
    {
        $person = $this->getDoctrine()->getManager()
            ->getRepository(\Chill\PersonBundle\Entity\Person::class)
            ->find($person_id);

        if (null === $person) {
            throw $this->createNotFoundException("Person with id {$person_id} not"
                    . ' found ');
        }

        $this->denyAccessUnlessGranted(
            'CHILL_PERSON_UPDATE',
            $person,
            'You are not allowed to edit this person.'
        );

        $address = new Address();

        $form = $this->createCreateForm($person, $address);

        return $this->render('ChillPersonBundle:Address:new.html.twig', [
            'person' => $person,
            'form' => $form->createView(),
        ]);
    }

    public function updateAction($person_id, $address_id, Request $request)
    {
        $person = $this->getDoctrine()->getManager()
            ->getRepository(\Chill\PersonBundle\Entity\Person::class)
            ->find($person_id);

        if (null === $person) {
            throw $this->createNotFoundException("Person with id {$person_id} not"
                . ' found ');
        }

        $this->denyAccessUnlessGranted(
            'CHILL_PERSON_UPDATE',
            $person,
            'You are not allowed to edit this person.'
        );

        $address = $this->findAddressById($person, $address_id);

        $form = $this->createEditForm($person, $address);
        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid()) {
            $validatePersonErrors = $this->validatePerson($person);

            if (count($validatePersonErrors) !== 0) {
                foreach ($validatePersonErrors as $error) {
                    $this->addFlash('error', $error->getMessage());
                }
            } elseif ($form->isValid()) {
                $this->getDoctrine()->getManager()
                    ->flush();

                $this->addFlash('success', $this->get('translator')->trans(
                    'The address has been successfully updated'
                ));

                return $this->redirectToRoute('chill_person_address_list', [
                    'person_id' => $person->getId(),
                ]);
            } else {
                $this->addFlash('error', $this->get('translator')
                    ->trans('Error when updating the period'));
            }
        }

        return $this->render('ChillPersonBundle:Address:edit.html.twig', [
            'person' => $person,
            'address' => $address,
            'form' => $form->createView(),
        ]);
    }

    /**
     * @return \Symfony\Component\Form\Form
     */
    protected function createCreateForm(Person $person, Address $address)
    {
        $form = $this->createForm(AddressType::class, $address, [
            'method' => 'POST',
            'action' => $this->generateUrl('chill_person_address_create', [
                'person_id' => $person->getId(),
            ]),
            'has_no_address' => true,
        ]);

        $form->add('submit', SubmitType::class, [
            'label' => 'Submit',
        ]);

        return $form;
    }

    /**
     * @return \Symfony\Component\Form\Form
     */
    protected function createEditForm(Person $person, Address $address)
    {
        $form = $this->createForm(AddressType::class, $address, [
            'method' => 'POST',
            'action' => $this->generateUrl('chill_person_address_update', [
                'person_id' => $person->getId(),
                'address_id' => $address->getId(),
            ]),
            'has_no_address' => true,
        ]);

        $form->add('submit', SubmitType::class, [
            'label' => 'Submit',
        ]);

        return $form;
    }

    /**
     * @param int $address_id
     *
     * @throws \Symfony\Component\HttpKernel\Exception\NotFoundHttpException if the address id does not exists or is not associated with given person
     *
     * @return Address
     */
    protected function findAddressById(Person $person, $address_id)
    {
        $address = $this->getDoctrine()->getManager()
            ->getRepository(Address::class)
            ->find($address_id);

        if (!$person->getAddresses()->contains($address)) {
            throw $this->createAccessDeniedException('Not allowed to see this address');
        }

        return $address;
    }

    /**
     * @param Chill\PersonBundle\Entity\Person $person
     *
     * @return \Symfony\Component\Validator\ConstraintViolationListInterface
     */
    private function validatePerson(Person $person)
    {
        $errors = $this->validator
            ->validate($person, null, ['Default']);
        $errors_addresses_consistent = $this->validator
            ->validate($person, null, ['addresses_consistent']);

        foreach ($errors_addresses_consistent as $error) {
            $errors->add($error);
        }

        return $errors;
    }
}
