<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\PersonBundle\Export\Export;

use Chill\CustomFieldsBundle\CustomFields\CustomFieldChoice;
use Chill\CustomFieldsBundle\Entity\CustomField;
use Chill\CustomFieldsBundle\Service\CustomFieldProvider;
use Chill\MainBundle\Export\ExportElementValidatedInterface;
use Chill\MainBundle\Export\FormatterInterface;
use Chill\MainBundle\Export\ListInterface;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Export\Declarations;
use Chill\PersonBundle\Security\Authorization\PersonVoter;
use DateTime;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Query;
use Exception;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\Extension\Core\Type\DateType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Security\Core\Role\Role;
use Symfony\Component\Validator\Constraints\Callback;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use Symfony\Contracts\Translation\TranslatorInterface;

use function addcslashes;
use function array_key_exists;
use function array_keys;
use function array_merge;
use function count;
use function in_array;
use function strtolower;
use function uniqid;

/**
 * Render a list of people.
 */
class ListPerson implements ExportElementValidatedInterface, ListInterface
{
    protected CustomFieldProvider $customFieldProvider;

    protected EntityManagerInterface $entityManager;

    protected array $fields = [
        'id',  'firstName', 'lastName', 'birthdate',
        'placeOfBirth', 'gender', 'memo', 'email', 'phonenumber',
        'mobilenumber', 'contactInfo', 'countryOfBirth', 'nationality',
        'address_street_address_1', 'address_street_address_2',
        'address_valid_from', 'address_postcode_label', 'address_postcode_code',
        'address_country_name', 'address_country_code', 'address_isnoaddress',
    ];

    protected TranslatableStringHelper $translatableStringHelper;

    protected TranslatorInterface $translator;

    private $slugs = [];

    public function __construct(
        EntityManagerInterface $em,
        TranslatorInterface $translator,
        TranslatableStringHelper $translatableStringHelper,
        CustomFieldProvider $customFieldProvider
    ) {
        $this->entityManager = $em;
        $this->translator = $translator;
        $this->translatableStringHelper = $translatableStringHelper;
        $this->customFieldProvider = $customFieldProvider;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $choices = array_combine($this->fields, $this->fields);

        foreach ($this->getCustomFields() as $cf) {
            $choices[$this->translatableStringHelper->localize($cf->getName())]
                =
                $cf->getSlug();
        }

        // Add a checkbox to select fields
        $builder->add('fields', ChoiceType::class, [
            'multiple' => true,
            'expanded' => true,
            'choices' => $choices,
            'label' => 'Fields to include in export',
            'choice_attr' => static function (string $val): array {
                // add a 'data-display-target' for address fields
                if (substr($val, 0, 8) === 'address_') {
                    return ['data-display-target' => 'address_date'];
                }

                return [];
            },
            'constraints' => [new Callback([
                'callback' => static function ($selected, ExecutionContextInterface $context) {
                    if (count($selected) === 0) {
                        $context->buildViolation('You must select at least one element')
                            ->atPath('fields')
                            ->addViolation();
                    }
                },
            ])],
        ]);

        // add a date  field for addresses
        $builder->add('address_date', DateType::class, [
            'label' => 'Address valid at this date',
            'data' => new DateTime(),
            'attr' => ['class' => 'datepicker'],
            'widget' => 'single_text',
            'format' => 'dd-MM-yyyy',
            'required' => false,
            'block_name' => 'list_export_form_address_date',
        ]);
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_LIST];
    }

    public function getDescription()
    {
        return 'Create a list of people according to various filters.';
    }

    public function getLabels($key, array $values, $data)
    {
        switch ($key) {
            case 'birthdate':
                // for birthdate, we have to transform the string into a date
                // to format the date correctly.
                return static function ($value) {
                    if ('_header' === $value) {
                        return 'birthdate';
                    }

                    if (empty($value)) {
                        return '';
                    }

                    $date = DateTime::createFromFormat('Y-m-d', $value);
                    // check that the creation could occurs.
                    if (false === $date) {
                        throw new Exception(sprintf('The value %s could '
                            . 'not be converted to %s', $value, DateTime::class));
                    }

                    return $date->format('d-m-Y');
                };

            case 'gender':
                // for gender, we have to translate men/women statement
                return function ($value) {
                    if ('_header' === $value) {
                        return 'gender';
                    }

                    return $this->translator->trans($value);
                };

            case 'countryOfBirth':
            case 'nationality':
                $countryRepository = $this->entityManager
                    ->getRepository(\Chill\MainBundle\Entity\Country::class);

                // load all countries in a single query
                 $countryRepository->findBy(['countryCode' => $values]);

                return function ($value) use ($key, $countryRepository) {
                    if ('_header' === $value) {
                        return strtolower($key);
                    }

                    if (null === $value) {
                        return $this->translator->trans('no data');
                    }

                    $country = $countryRepository->find($value);

                    return $this->translatableStringHelper->localize(
                        $country->getName()
                    );
                };

            case 'address_country_name':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return strtolower($key);
                    }

                    if (null === $value) {
                        return '';
                    }

                    return $this->translatableStringHelper->localize(json_decode($value, true));
                };

            case 'address_isnoaddress':
                return static function (?string $value): string {
                    if ('_header' === $value) {
                        return 'address.address_homeless';
                    }

                    if (null !== $value) {
                        return 'X';
                    }

                    return '';
                };

            default:
                // for fields which are associated with person
                if (in_array($key, $this->fields, true)) {
                    return static function ($value) use ($key) {
                        if ('_header' === $value) {
                            return strtolower($key);
                        }

                        return $value;
                    };
                }

                return $this->getLabelForCustomField($key, $values, $data);
        }
    }

    public function getQueryKeys($data)
    {
        $fields = [];

        foreach ($data['fields'] as $key) {
            if (in_array($key, $this->fields, true)) {
                $fields[] = $key;
            }
        }

        // add the key from slugs and return
        return array_merge($fields, array_keys($this->slugs));
    }

    public function getResult($query, $data)
    {
        return $query->getQuery()->getResult(Query::HYDRATE_SCALAR);
    }

    public function getTitle()
    {
        return 'List peoples';
    }

    public function getType()
    {
        return Declarations::PERSON_TYPE;
    }

    public function initiateQuery(array $requiredModifiers, array $acl, array $data = [])
    {
        $centers = array_map(static function ($el) {
            return $el['center'];
        }, $acl);

        // throw an error if any fields are present
        if (!array_key_exists('fields', $data)) {
            throw new \Doctrine\DBAL\Exception\InvalidArgumentException('any fields '
                . 'have been checked');
        }

        $qb = $this->entityManager->createQueryBuilder();

        foreach ($this->fields as $f) {
            if (in_array($f, $data['fields'], true)) {
                switch ($f) {
                    case 'countryOfBirth':
                    case 'nationality':
                        $qb->addSelect(sprintf('IDENTITY(person.%s) as %s', $f, $f));

                        break;

                    case 'address_street_address_1':
                    case 'address_street_address_2':
                    case 'address_valid_from':
                    case 'address_postcode_label':
                    case 'address_postcode_code':
                    case 'address_country_name':
                    case 'address_country_code':
                    case 'address_isnoaddress':
                        $qb->addSelect(sprintf(
                            'GET_PERSON_ADDRESS_%s(person.id, :address_date) AS %s',
                            // get the part after address_
                            strtoupper(substr($f, 8)),
                            $f
                        ));
                        $qb->setParameter('address_date', $data['address_date']);

                        break;

                    default:
                        $qb->addSelect(sprintf('person.%s as %s', $f, $f));
                }
            }
        }

        foreach ($this->getCustomFields() as $cf) {
            $cfType = $this->customFieldProvider->getCustomFieldByType($cf->getType());

            if ($cfType instanceof CustomFieldChoice && $cfType->isMultiple($cf)) {
                foreach ($cfType->getChoices($cf) as $choiceSlug => $label) {
                    $slug = $this->slugToDQL($cf->getSlug(), 'choice', ['choiceSlug' => $choiceSlug]);
                    $qb->addSelect(
                        sprintf(
                            'GET_JSON_FIELD_BY_KEY(person.cFData, :slug%s) AS %s',
                            $slug,
                            $slug
                        )
                    );
                    $qb->setParameter(sprintf('slug%s', $slug), $cf->getSlug());
                }
            } else {
                $slug = $this->slugToDQL($cf->getSlug());
                $qb->addSelect(
                    sprintf(
                        'GET_JSON_FIELD_BY_KEY(person.cFData, :slug%s) AS %s',
                        $slug,
                        $slug
                    )
                );
                $qb->setParameter(sprintf('slug%s', $slug), $cf->getSlug());
            }
        }

        $qb
            ->from('ChillPersonBundle:Person', 'person')
            ->join('person.center', 'center')
            ->andWhere('center IN (:authorized_centers)')
            ->setParameter('authorized_centers', $centers);

        return $qb;
    }

    public function requiredRole()
    {
        return new Role(PersonVoter::LISTS);
    }

    public function supportsModifiers()
    {
        return [Declarations::PERSON_TYPE, Declarations::PERSON_IMPLIED_IN];
    }

    public function validateForm($data, ExecutionContextInterface $context)
    {
        // get the field starting with address_
        $addressFields = array_filter(
            $this->fields,
            static fn (string $el): bool => substr($el, 0, 8) === 'address_'
        );

        // check if there is one field starting with address in data
        if (count(array_intersect($data['fields'], $addressFields)) > 0) {
            // if a field address is checked, the date must not be empty
            if (empty($data['address_date'])) {
                $context
                    ->buildViolation('You must set this date if an address is checked')
                    ->atPath('address_date')
                    ->addViolation();
            }
        }
    }

    private function DQLToSlug($cleanedSlug)
    {
        return $this->slugs[$cleanedSlug]['slug'];
    }

    /**
     * @param mixed $slug
     *
     * @return array An array with keys = 'slug', 'type', 'additionnalInfo'
     */
    private function extractInfosFromSlug($slug): array
    {
        return $this->slugs[$slug];
    }

    /**
     * Get custom fields associated with person.
     *
     * @return CustomField[]
     */
    private function getCustomFields()
    {
        return $this->entityManager
            ->createQuery('SELECT cf '
                . 'FROM ChillCustomFieldsBundle:CustomField cf '
                . 'JOIN cf.customFieldGroup g '
                . 'WHERE cf.type != :title AND g.entity LIKE :entity')
            ->setParameters([
                'title' => 'title',
                'entity' => addcslashes(Person::class, '\\'),
            ])
            ->getResult();
    }

    private function getLabelForCustomField($key, array $values, $data)
    {
        // for fields which are custom fields
        /** @var CustomField $cf */
        $cf = $this->entityManager
            ->getRepository(CustomField::class)
            ->findOneBy(['slug' => $this->DQLToSlug($key)]);
        $cfType = $this->customFieldProvider->getCustomFieldByType($cf->getType());
        $defaultFunction = function ($value) use ($cf) {
            if ('_header' === $value) {
                return $this->translatableStringHelper->localize($cf->getName());
            }

            return $this->customFieldProvider
                ->getCustomFieldByType($cf->getType())
                ->render(json_decode($value, true), $cf, 'csv');
        };

        if ($cfType instanceof CustomFieldChoice && $cfType->isMultiple($cf)) {
            return function ($value) use ($cf, $cfType, $key) {
                $slugChoice = $this->extractInfosFromSlug($key)['additionnalInfos']['choiceSlug'];
                $decoded = json_decode($value, true);

                if ('_header' === $value) {
                    $label = $cfType->getChoices($cf)[$slugChoice];

                    return $this->translatableStringHelper->localize($cf->getName())
                            . ' | ' . $label;
                }

                if ('_other' === $slugChoice && $cfType->isChecked($cf, $choiceSlug, $decoded)) {
                    return $cfType->extractOtherValue($cf, $decoded);
                }

                return $cfType->isChecked($cf, $slugChoice, $decoded);
            };
        }

        return $defaultFunction;
    }

    /**
     * Clean a slug to be usable by DQL.
     */
    private function slugToDQL(string $slug, string $type = 'default', array $additionalInfos = []): string
    {
        $uid = 'slug_' . uniqid('', true);

        $this->slugs[$uid] = [
            'slug' => $slug,
            'type' => $type,
            'additionnalInfos' => $additionalInfos,
        ];

        return $uid;
    }
}
