<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\PersonBundle\Repository;

use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Security\Authorization\AuthorizationHelper;
use Chill\MainBundle\Security\Resolver\CenterResolverDispatcherInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Entity\Person;
use DateTime;

use Symfony\Component\Security\Core\Security;
use function count;

final class AccompanyingPeriodACLAwareRepository implements AccompanyingPeriodACLAwareRepositoryInterface
{
    private AccompanyingPeriodRepository $accompanyingPeriodRepository;

    private AuthorizationHelper $authorizationHelper;

    private CenterResolverDispatcherInterface $centerResolverDispatcher;

    private Security $security;

    public function __construct(
        AccompanyingPeriodRepository $accompanyingPeriodRepository,
        Security $security,
        AuthorizationHelper $authorizationHelper,
        CenterResolverDispatcherInterface $centerResolverDispatcher
    ) {
        $this->accompanyingPeriodRepository = $accompanyingPeriodRepository;
        $this->security = $security;
        $this->authorizationHelper = $authorizationHelper;
        $this->centerResolverDispatcher = $centerResolverDispatcher;
    }

    public function buildQueryOpenedAccompanyingCourseByUser(?User $user)
    {
        $qb = $this->accompanyingPeriodRepository->createQueryBuilder('ap');

        $qb->where($qb->expr()->eq('ap.user', ':user'))
            ->andWhere(
                $qb->expr()->neq('ap.step', ':draft'),
                $qb->expr()->orX(
                    $qb->expr()->isNull('ap.closingDate'),
                    $qb->expr()->gt('ap.closingDate', ':now')
                )
            )
            ->setParameter('user', $user)
            ->setParameter('now', new DateTime('now'))
            ->setParameter('draft', AccompanyingPeriod::STEP_DRAFT);

        return $qb;
    }

    public function countByUserOpenedAccompanyingPeriod(?User $user): int
    {
        if (null === $user) {
            return 0;
        }

        return $this->buildQueryOpenedAccompanyingCourseByUser($user)
            ->select('COUNT(ap)')
            ->getQuery()
            ->getSingleScalarResult();
    }

    public function findByPerson(
        Person $person,
        string $role,
        ?array $orderBy = [],
        ?int $limit = null,
        ?int $offset = null
    ): array {
        $qb = $this->accompanyingPeriodRepository->createQueryBuilder('ap');
        $scopes = $this->authorizationHelper
            ->getReachableCircles(
                $this->security->getUser(),
                $role,
                $this->centerResolverDispatcher->resolveCenter($person)
            );

        if (0 === count($scopes)) {
            return [];
        }

        $qb
            ->join('ap.participations', 'participation')
            ->where($qb->expr()->eq('participation.person', ':person'))
            ->andWhere(
                $qb->expr()->orX(
                    'ap.confidential = FALSE',
                    $qb->expr()->eq('ap.user', ':user')
                )
            )
            ->andWhere(
                $qb->expr()->orX(
                    $qb->expr()->neq('ap.step', ':draft'),
                    $qb->expr()->eq('ap.createdBy', ':creator')
                )
            )
            ->setParameter('draft', AccompanyingPeriod::STEP_DRAFT)
            ->setParameter('person', $person)
            ->setParameter('user', $this->security->getUser())
            ->setParameter('creator', $this->security->getUser());
        // add join condition for scopes
        $orx = $qb->expr()->orX(
            $qb->expr()->eq('ap.step', ':draft')
        );

        foreach ($scopes as $key => $scope) {
            $orx->add($qb->expr()->isMemberOf(':scope_' . $key, 'ap.scopes'));
            $qb->setParameter('scope_' . $key, $scope);
        }
        $qb->andWhere($orx);

        return $qb->getQuery()->getResult();
    }

    /**
     * @return array|AccompanyingPeriod[]
     */
    public function findByUserOpenedAccompanyingPeriod(?User $user, array $orderBy = [], int $limit = 0, int $offset = 50): array
    {
        if (null === $user) {
            return [];
        }

        $qb = $this->buildQueryOpenedAccompanyingCourseByUser($user);

        $qb->setFirstResult($offset)
            ->setMaxResults($limit);

        foreach ($orderBy as $field => $direction) {
            $qb->addOrderBy('ap.' . $field, $direction);
        }

        return $qb->getQuery()->getResult();
    }
}
