<?php

/**
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

declare(strict_types=1);

namespace Chill\PersonBundle\Security\Authorization;

use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Security\Authorization\AbstractChillVoter;
use Chill\MainBundle\Security\Authorization\VoterHelperFactoryInterface;
use Chill\MainBundle\Security\Authorization\VoterHelperInterface;
use Chill\MainBundle\Security\ProvideRoleHierarchyInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Entity\Person;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Symfony\Component\Security\Core\Security;

use function in_array;

class AccompanyingPeriodVoter extends AbstractChillVoter implements ProvideRoleHierarchyInterface
{
    /**
     * all the roles that are linked to an accompanying period.
     */
    public const ALL = [
        self::SEE,
        self::SEE_DETAILS,
        self::CREATE,
        self::EDIT,
        self::DELETE,
        self::FULL,
        self::TOGGLE_CONFIDENTIAL_ALL,
        self::TOGGLE_INTENSITY,
        self::RE_OPEN_COURSE,
    ];

    /**
     * Give the ability to see all confidential courses.
     */
    public const CONFIDENTIAL_CRUD = 'CHILL_PERSON_ACCOMPANYING_PERIOD_CRUD_CONFIDENTIAL';

    public const CREATE = 'CHILL_PERSON_ACCOMPANYING_PERIOD_CREATE';

    /**
     * role to DELETE the course.
     *
     * Will be true only for the creator, and if the course is still at DRAFT step.
     */
    public const DELETE = 'CHILL_PERSON_ACCOMPANYING_PERIOD_DELETE';

    /**
     * role to EDIT the course.
     *
     * If the course is closed, it will be always false.
     */
    public const EDIT = 'CHILL_PERSON_ACCOMPANYING_PERIOD_UPDATE';

    /**
     * Give all the right above.
     */
    public const FULL = 'CHILL_PERSON_ACCOMPANYING_PERIOD_FULL';

    /**
     * Reopen a closed course.
     *
     * This forward to the EDIT role, without taking into account that the course
     * is closed
     */
    public const RE_OPEN_COURSE = 'CHILL_PERSON_ACCOMPANYING_PERIOD_REOPEN';

    /**
     * Allow user to bulk reassign the courses.
     */
    public const REASSIGN_BULK = 'CHILL_PERSON_ACCOMPANYING_COURSE_REASSIGN_BULK';

    public const SEE = 'CHILL_PERSON_ACCOMPANYING_PERIOD_SEE';

    /**
     * details are for seeing:.
     *
     * * SocialIssues
     */
    public const SEE_DETAILS = 'CHILL_PERSON_ACCOMPANYING_PERIOD_SEE_DETAILS';

    public const TOGGLE_CONFIDENTIAL = 'CHILL_PERSON_ACCOMPANYING_PERIOD_TOGGLE_CONFIDENTIAL';

    /**
     * Right to toggle confidentiality.
     */
    public const TOGGLE_CONFIDENTIAL_ALL = 'CHILL_PERSON_ACCOMPANYING_PERIOD_TOGGLE_CONFIDENTIAL_ALL';

    /**
     * Right to toggle urgency of parcours.
     */
    public const TOGGLE_INTENSITY = 'CHILL_PERSON_ACCOMPANYING_PERIOD_TOGGLE_INTENSITY';

    private Security $security;

    private VoterHelperInterface $voterHelper;

    public function __construct(
        Security $security,
        VoterHelperFactoryInterface $voterHelperFactory
    ) {
        $this->security = $security;
        $this->voterHelper = $voterHelperFactory
            ->generate(self::class)
            ->addCheckFor(null, [self::CREATE, self::REASSIGN_BULK])
            ->addCheckFor(AccompanyingPeriod::class, [self::TOGGLE_CONFIDENTIAL, ...self::ALL])
            ->addCheckFor(Person::class, [self::SEE])
            ->build();
    }

    public function getRoles(): array
    {
        return [
            self::SEE,
            self::SEE_DETAILS,
            self::CONFIDENTIAL_CRUD,
            self::CREATE,
            self::EDIT,
            self::DELETE,
            self::FULL,
            self::TOGGLE_CONFIDENTIAL_ALL,
            self::REASSIGN_BULK,
        ];
    }

    public function getRolesWithHierarchy(): array
    {
        return ['Accompanying period' => $this->getRoles()];
    }

    public function getRolesWithoutScope(): array
    {
        return [self::REASSIGN_BULK];
    }

    protected function supports($attribute, $subject)
    {
        return $this->voterHelper->supports($attribute, $subject);
    }

    protected function voteOnAttribute($attribute, $subject, TokenInterface $token): bool
    {
        if (!$token->getUser() instanceof User) {
            return false;
        }

        if ($subject instanceof AccompanyingPeriod) {
            if (AccompanyingPeriod::STEP_CLOSED === $subject->getStep()) {
                if (in_array($attribute, [self::EDIT, self::DELETE], true)) {
                    return false;
                }

                if (self::RE_OPEN_COURSE === $attribute) {
                    return $this->voterHelper->voteOnAttribute(self::EDIT, $subject, $token);
                }
            }

            if (AccompanyingPeriod::STEP_DRAFT === $subject->getStep()) {
                // only creator can see, edit, delete, etc.
                if (
                    $subject->getCreatedBy() === $token->getUser()
                    || null === $subject->getCreatedBy()
                ) {
                    return true;
                }

                return false;
            }

            if (in_array($attribute, [
                self::SEE, self::SEE_DETAILS, self::EDIT,
            ], true)) {
                if ($subject->getUser() === $token->getUser()) {
                    return true;
                }
            }

            if (self::TOGGLE_CONFIDENTIAL === $attribute) {
                if (null !== $subject->getUser() && ($subject->getUser() === $token->getUser())) {
                    return true;
                }

                if ($this->voterHelper->voteOnAttribute(self::TOGGLE_CONFIDENTIAL_ALL, $subject, $token)) {
                    return true;
                }

                return false;
            }

            if (self::TOGGLE_INTENSITY === $attribute) {
                if (null !== $subject->getUser() && ($subject->getUser() === $token->getUser())) {
                    return true;
                }

                return false;
            }

            // if confidential, only the referent can see it
            if ($subject->isConfidential()) {
                if ($this->voterHelper->voteOnAttribute(self::CONFIDENTIAL_CRUD, $subject, $token)) {
                    return true;
                }

                return $token->getUser() === $subject->getUser();
            }
        }

        return $this->voterHelper->voteOnAttribute($attribute, $subject, $token);
    }
}
